# frozen_string_literal: true

# Released under the MIT License.
# Copyright, 2012-2013, by Tony Arcieri.
# Copyright, 2015-2024, by Samuel Williams.

module Async
	module DNS
		VERSION = "1.4.1"
	end
end
